import logging
import re

logger = logging.getLogger(__name__)


migration_tests = (
    {
        "code": "NOT_NULL",
        "fn": lambda sql, **kw: re.search("NOT NULL", sql)
        and not re.search("CREATE TABLE", sql),
        "err_msg": "NOT NULL constraint on columns",
    },
    {
        "code": "DROP_COLUMN",
        "fn": lambda sql, **kw: re.search("DROP COLUMN", sql),
        "err_msg": "DROPPING columns",
    },
    {
        "code": "RENAME_COLUMN",
        "fn": lambda sql, **kw: re.search("ALTER TABLE .* CHANGE", sql)
        or re.search("ALTER TABLE .* RENAME COLUMN", sql),
        "err_msg": "RENAMING columns",
    },
    {
        "code": "RENAME_TABLE",
        "fn": lambda sql, **kw: re.search("RENAME TABLE", sql)
        or re.search("ALTER TABLE .* RENAME TO", sql),
        "err_msg": "RENAMING tables",
    },
    {
        "code": "DROP_TABLE",
        "fn": lambda sql, **kw: re.search("DROP TABLE", sql),
        "err_msg": "DROPPING tables",
    },
    {
        "code": "CREATE_INDEX_SYNC",
        "fn": lambda sql, **kw: re.search("CREATE INDEX", sql) and not re.search("CONCURRENTLY", sql),
        "err_msg": "CREATING INDEX SYNC",
    },
    {
        "code": "DROP_INDEX_SYNC",
        "fn": lambda sql, **kw: re.search("DROP INDEX", sql) and not re.search("CONCURRENTLY", sql),
        "err_msg": "DROP INDEX SYNC",
    },
    {
        "code": "ALTER_COLUMN",
        "fn": lambda sql, **kw: re.search("ALTER TABLE .* MODIFY", sql)
        or re.search("ALTER TABLE .* ALTER COLUMN .* TYPE", sql),
        "err_msg": (
            "ALTERING columns (Could be backwards incompatible. "
            "Check operation to be sure.)"
        ),
    },
)


def analyse_sql_statements(sql_statements):
    errors = []
    for statement in sql_statements:
        for test in migration_tests:
            if test["fn"](statement, errors=errors):
                logger.debug("Testing {0} -- ERROR".format(statement))
                table_search = re.search("TABLE `([^`]*)`", statement, re.IGNORECASE)
                col_search = re.search("COLUMN `([^`]*)`", statement, re.IGNORECASE)
                err = {
                    "err_msg": test["err_msg"],
                    "code": test["code"],
                    "table": table_search.group(1) if table_search else None,
                    "column": col_search.group(1) if col_search else None,
                }
                errors.append(err)
            else:
                logger.debug("Testing {0} -- PASSED".format(statement))
    return errors



from __future__ import print_function

import hashlib
import logging
import os
import sys

from django.core.management import call_command

from .sql_analyser import analyse_sql_statements
from .state import State
from .utils import get_migration_abspath

logger = logging.getLogger(__name__)

DJANGO_APPS_WITH_MIGRATIONS = ("admin", "auth", "contenttypes", "sessions")


class MigrationLinter:

    def __init__(self, std_out=sys.stdout, state_path=".migration_state", include_apps=None, interactive=False, force_update=False):
        self.state = State(state_path)
        self.state.load()
        self.interactive = interactive
        self.force_update = force_update
        self.include_apps = include_apps
        self.stdout = std_out

    def write(self, text):
        self.stdout.write(text)

    def lint_all_migrations(self):
        migrations = self._gather_all_migrations()

        # Lint those migrations
        sorted_migrations = sorted(
            migrations, key=lambda migration: (migration.app_label, migration.name)
        )
        for m in sorted_migrations:
            if self.include_apps and m.app_label not in self.include_apps:
                continue

            success = self.lint_migration(m)
            if not success:
                return False
        if self.interactive:
            self.state.save()
        return True

    def lint_migration(self, migration):
        app_label = migration.app_label
        migration_name = migration.name

        migration_hash = self.get_migration_hash(app_label, migration_name)

        if (
            self.state and migration_name in self.state[app_label]
            and self.state[app_label][migration_name]["migration_hash"]
            == migration_hash
        ):
            return True

        self.write(f"({app_label}, {migration_name})... ")
        try:
            sql_statements = self.get_sql(app_label, migration_name)
        except Exception as e:
            errors = [
                {
                    "err_msg": str(e),
                    "code": "SQL_RETRIEVE_ERROR",
                    "table": None,
                    "column": None,
                }
            ]
        else:
            errors = analyse_sql_statements(sql_statements)

        forced = False
        if errors:
            self.write("ERR")
            self.print_errors(errors)

            if self.interactive:
                self.write("Invalid migration! It is backwards incompatible. Please update the migration file.")
                response = input('To edit the file, hit ENTER. Type "allow" if you feel this is incorrect": ')
                if response != "allow":
                    return False

                forced = True
            elif self.force_update:
                forced = True
            else:
                return False
        else:
            self.write("OK")

        if self.interactive:
            self.state[app_label][migration_name] = dict(
                migration_name=migration_name,
                migration_hash=migration_hash,
                migration_forced="migration_forced" if forced else "",
            )

        return True

    @staticmethod
    def get_migration_hash(app_label, migration_name):
        hash_md5 = hashlib.md5()
        with open(get_migration_abspath(app_label, migration_name), "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def print_errors(self, errors):
        for err in errors:
            error_str = "\t{0}".format(err["err_msg"])
            if err["table"]:
                error_str += " (table: {0}".format(err["table"])
                if err["column"]:
                    error_str += ", column: {0}".format(err["column"])
                error_str += ")"
            self.write(error_str)

    def get_sql(self, app_label, migration_name):
        dev_null = open(os.devnull, "w")
        sql_statement = call_command(
            "sqlmigrate", app_label, migration_name, stdout=dev_null
        )
        return sql_statement.splitlines()

    @staticmethod
    def _gather_all_migrations():
        from django.db.migrations.loader import MigrationLoader

        migration_loader = MigrationLoader(connection=None, load=False)
        migration_loader.load_disk()
        # Prune Django apps
        for (app_label, _), migration in migration_loader.disk_migrations.items():
            if app_label not in DJANGO_APPS_WITH_MIGRATIONS:
                yield migration



import sys

from django.core.management.base import BaseCommand

from ...migration_linter import MigrationLinter


class Command(BaseCommand):
    help = "Lint your migrations"

    def add_arguments(self, parser):
        parser.add_argument(
            "--interactive",
            dest="interactive",
            action="store_true",
            help=(
                "if specified, allows you to force a migration, even if it is determined to be backwards incompatible"
            ),
        )

        parser.add_argument(
            "--include-apps",
            nargs='+',
            dest="include_apps",
            help=("Apps to lint. If not included it will lint all"),
        )

        parser.add_argument(
            "--force-update",
            dest="force_update",
            action="store_true",
            help=(
                "overwrite existing state and use current state as source of truth."
                "This is only useful when first starting a project."
            ),
        )

        parser.add_argument(
            "--state-path",
            dest="state_path",
            default=".migration_state",
            help=(
                "The folder to store the state of all migrations"
            ),
        )

    def handle(self, *args, interactive, force_update, include_apps, state_path, **options):
        linter = MigrationLinter(
            std_out=self.stdout, state_path=state_path,
            interactive=interactive, include_apps=include_apps, force_update=force_update
        )
        success = linter.lint_all_migrations()

        if not success:
            self.stdout.write("Linting failed. Please check errors above")
            sys.exit(1)
        else:
            self.stdout.write("Linting complete.")



import base64
import csv
import os
from collections import defaultdict


class State(defaultdict):
    field_names = ["migration_name", "migration_hash", "migration_forced"]

    def __init__(self, cache_path):
        if not os.path.exists(cache_path):
            os.makedirs(cache_path)
        self.cache_path = cache_path
        super().__init__(dict)

    def load(self):
        self.clear()
        for file_name in os.listdir(self.cache_path):
            file_path = os.path.join(self.cache_path, file_name)
            if os.path.isfile(file_path):
                rows = []
                with open(file_path, "r") as csv_file:
                    reader = csv.DictReader(csv_file, fieldnames=self.field_names)
                    for row in reader:
                        rows.append(row)
                self[file_name] = {row["migration_name"]: row for row in rows}

    def save(self):
        for key in self:
            with open(os.path.join(self.cache_path, key), "w") as csv_file:
                writer = csv.DictWriter(csv_file, fieldnames=self.field_names)
                for row in sorted(
                    self[key].values(), key=lambda m: m["migration_name"]
                ):
                    writer.writerow(row)
